/* 
 * EDOMI beta (Version 1.52) - Intelligente Steuerung und Visualisierung KNX-basierter Elektro-Installationen.
 * @link http://www.edomi.de 
 * @author Dr. Christian Gärtner <dr.christian.gaertner@googlemail.com> 
 * @copyright 2015,2016,2017 Dr. Christian Gärtner 
 * @license Freeware 
 *
 *     This file and all other files are part of EDOMI.
 *
 *     EDOMI is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 *     Diese und alle anderen Dateien sind Teil von EDOMI.
 *
 *     EDOMI wird in der Hoffnung, dass es nützlich sein wird, aber
 *     OHNE JEDE GEWÄHRLEISTUNG, bereitgestellt; sogar ohne die implizite
 *     Gewährleistung der MARKTFÄHIGKEIT oder EIGNUNG FÜR EINEN BESTIMMTEN ZWECK.
*/ 
/*
============================================================================
Initialisierung
============================================================================
*/

function firstinit(visuid,newSid,version) {
	//dies wird NUR EINMALIG beim allerersten Start aufgerufen (also nach dem Laden der index.php)
	
	//--- ggf. anpassen ---
	displayPixelRatio=2;			//Faktor für die virtuelle Auflösung eines CANVAS (z.B. für Retina-Displays sollte der Faktor mindestens 2 betragen)
									//benötigt für: Diagramme, Rundinstrumente
	//---------------------

	sid=newSid;
	edomiVersion=version;

	visu_visuid=null;				//aktuelle VisuID (wird erst in app0.php/initVisu gesetzt)
	visu_startPageId=null;			//Startseite
	visu_sspageid=null;				//Bildschirmschoner-Seiten-ID
	visu_sstimeout=null; 			//nach dieser Zeit (Minuten!) wird der Bilderschirmschoner aufgerufen

	visu_indiColor='#80e000';		//Farbe für alle Indikatoren (Klicks, ZSU, etc.)
	visu_indiColorText='inherit';	//Text-Farbe für einige Indikatoren (Tastatureingabe, Drehregler, etc.)

	visu_touchMode=0;				//Touch oder Click? (0=Automatisch, 1=Klick, 2=Touch)
	visu_allowScrolling=false;		//iOS: Scrollen erlauben?

	visu_isMobile=(navigator.userAgent.match(/iPhone|iPad|iPod|iOS|Android/i)?true:false);	//true=Mobile-Device

	visu_sound=false;				//Audio-Objekt
	visu_soundEnabled=0;			//0=Sound wird nicht benötigt, 1=wird benötigt ist aber noch nicht initialisiert (ggf. iOS), 2=steht zu Verfügung und ist initialisiert
	visu_textToSpeech=false; 		//SpeechSynthesisUtterance-Objekt
	visu_textToSpeechEnabled=0;		//0=TextToSpeech wird nicht benötigt, 1=wird benötigt ist aber noch nicht initialisiert (iOS), 2=steht zu Verfügung und ist initialisiert

	preload_images_max=0;			//Preload-Bilder (Anzahl, nur temporär für den Preloadvorgang!)
	preload_imagesFile=new Array();	//Preload-Bilder (Dateinamen, nur temporär für den Preloadvorgang!)

	visu_socket=null;

	initVisu(visuid);
}

function initVisu(visuid) {
	//wird beim Start und Fehler (=Neustart) aufgerufen

	windowZindex=0; 				//Fenster-Verwaltung

	errorTimer=null; 				//Timer für Fehlerbehandlung (serverNotReady)
	screensaverTimer=null; 			//Timer für Bildschirmschoner
	clickTimer=null; 				//Timer für Spezial-Click

	screensaverActive=false;
	initScreensaverTimer();			//Bildschirmschoner zurücksetzen

	elementDragId=0;				//für Controltypen mit Drag-Funktion
	
	visu_showWarningErr=0;			//Fehlerhinweis quittieren: Enthält die Anzahl der aktuell quittierten Fehler, -1=deaktiviert

	ajax("initVisu",0,visuid,0,"","");
}

function prepareVisu() {
	document.getElementById("waitanim").style.borderLeftColor="#20e000";
	document.getElementById("waitanim").style.borderRightColor="#20e000";
	if (preload_imagesFile.length>0) {
		document.getElementById("preload").style.display="block";
		preload_images_max=preload_imagesFile.length;
		preloadImagesLoad();
	} else {
		startVisu();
	}
}

function preloadImagesLoad() {
	if (preload_imagesFile.length>0) {
		var id=preload_imagesFile[0].match(/visu\/img\/img-([0-9]*)./);
		if (id && id[1]>0) {
			var obj=new Image();			
			obj.onload=function(){preloadImagesReady();}
			obj.onerror=function(){preloadImagesReady();}
			obj.src=preload_imagesFile.shift();
		} else {
			preload_imagesFile.shift();
			preloadImagesReady();		
		}
	} else {
		startVisu();
	}
}

function preloadImagesReady() {
	var t=parseInt((preload_images_max-preload_imagesFile.length)*100/preload_images_max);
	document.getElementById("preload").style.background="-webkit-linear-gradient(left,#ffffff 0%,#ffffff "+t+"%,#606060 "+t+"%,#606060 100%)";
	preloadImagesLoad();
}

function startVisu() {
	document.getElementById("waitanim").style.borderLeftColor="#e00000";
	document.getElementById("waitanim").style.borderRightColor="#e00000";
	document.getElementById("wait").style.display="none";
	document.getElementById("preload").style.display="none";
	document.getElementById("windowContainer").style.display="block";
	openPage(visu_startPageId);
}

function jsLogin(visuid,visus) {
	//Logindialog (modal)	
	gotoStart();
	var loginWinId="login";
	var n="<div id='loginform' class='appWindowLogin' onkeydown='if (event.keyCode==13) {ajax(\"login\",0,\"\",\"\",\"\",controlGetFormData(\"loginform\"));}' style='width:200px;'>";
		n+="<div style='margin-bottom:10px;'><b>Login</b></div>";
		n+="<select id='"+loginWinId+"-loginform-fd0' class='loginSelect' data-type='1' style='margin-bottom:15px;'>"+visus+"</select>";
		n+="<input type='text' id='"+loginWinId+"-loginform-fd1' data-type='1' value='' maxlength='30' class='loginInput' autofocus style='width:100%; height:25px; padding:5px; border-color:transparent;'></input><br>";
		n+="<input type='password' id='"+loginWinId+"-loginform-fd2' data-type='1' value='' maxlength='30' class='loginInput' style='width:100%; height:25px; padding:5px; border-color:transparent;'></input><br>";
		n+="<br>";
		n+="<button id='"+loginWinId+"-loginform-fd3' class='loginCheckbox' onClick='checkboxToggle(this);' data-type='5' data-list='normal|Neuladen' data-value='"+((visu_isMobile)?"1":"0")+"' style='float:left;'></button>";
		n+="<div class='loginButton' onClick='ajax(\"login\",0,\"\",\"\",\"\",controlGetFormData(\"loginform\"));' style='float:right; border-color:transparent;'>Ok</div>";
	n+="</div>";
	
	//POST-Login (für alterative Login-Methode unter iOS & Co.)
	n+="<form id='loginform-post' action='index.php' method='post'><input type='hidden' id='loginvisuid' name='loginvisuid' value=''></input><input type='hidden' id='loginsid' name='loginsid' value=''></input></form>";

	document.getElementById(loginWinId).innerHTML=n;
	checkboxInit(document.getElementById(loginWinId+"-loginform-fd3"));
	document.getElementById(loginWinId).style.display="block";
}


/*
============================================================================
Sound & Speech
============================================================================
*/

function visuSoundInit(obj) {
	if (visu_soundEnabled==1) {
		if (!visu_sound) {visu_sound=new Audio("../shared/etc/snd-empty.mp3");}	//Dummy
		visu_sound.type="audio/mpeg";
		visu_sound.preload="none";
		visu_sound.load();
		visu_sound.autoplay=false;
		visu_sound.loop=false;
		visu_sound.onplay=function(event){
			visu_soundEnabled=2;
			if (obj && visu_soundEnabled!=1 && visu_textToSpeechEnabled!=1) {obj.style.display="none";}
		}
		visu_sound.play();
	}
}

function visuSoundPlay(sndId,sndUrl,sndLoop) {
	if (visu_sound && visu_soundEnabled==2) {
		if (sndId>0) {sndUrl="../data/liveproject/visu/etc/snd-"+sndId+".mp3";}
		visu_sound.src=sndUrl; 
		visu_sound.load();
		visu_sound.loop=sndLoop;
		visu_sound.play();
	}
}

function visuSoundStop() {
	if (visu_sound && visu_soundEnabled==2) {
		visu_sound.src="../shared/etc/snd-empty.mp3";
		visu_sound.load();
		visu_sound.pause();
	}
}

function visuTextToSpeechInit(obj) {
	if (visu_textToSpeechEnabled==1 && ("speechSynthesis" in window)) {
		if (!visu_textToSpeech) {visu_textToSpeech=new SpeechSynthesisUtterance();}
		visu_textToSpeech.onstart=function(event){
			visu_textToSpeechEnabled=2;
			if (obj && visu_soundEnabled!=1 && visu_textToSpeechEnabled!=1) {obj.style.display="none";}
		}
		visu_textToSpeech.text=" ";	//Dummy (darf nicht leer sein, muss bei Init ausgesprochen werden!)
		window.speechSynthesis.speak(visu_textToSpeech);
	}
}	

function visuTextToSpeechPlay(voice,language,rate,pitch,msg) {
	if (visu_textToSpeech && visu_textToSpeechEnabled==2 && msg!="") {
		visu_textToSpeech.text=msg;
		if (voice!="") 		{visu_textToSpeech.voice=voice;}	//### Voice wird noch nicht wirklich unterstützt...
		if (language!="") 	{visu_textToSpeech.lang=language;}
		if (rate!="") 		{visu_textToSpeech.rate=rate;}
		if (pitch!="") 		{visu_textToSpeech.pitch=pitch;}
		window.speechSynthesis.speak(visu_textToSpeech);
	}
}

function visuTextToSpeechStop() {
	if (visu_textToSpeech && visu_textToSpeechEnabled==2) {
		window.speechSynthesis.cancel();
	}
}


/*
============================================================================
Visuseiten & Popups
============================================================================
*/

function openPage(pageid,sspage) {
	//Visuseite/Popup aufrufen
	cancelAll();

	//Bildschirmschoner-Reset (sofern nicht die BS-Seite selbst aufgerufen wird)
	if (sspage!==true) {initScreensaverTimer();}

	visu_socket.request_initPage(pageid);
}

function reloadPage() {
	//### ungenutzt
	//aktuelle Visuseite (nicht Popup!) neu laden (KOs updaten)
	cancelAll();
	var obj=document.getElementById("w0");
	if (obj && obj.dataset.pageid>0) {
		openPage(obj.dataset.pageid);
		visu_socket.request_initPage(pageid);
	}
}

function focusPopup(pageid) {
	//existiert bereits ein Popup(!) mit dieser pageid, wird das Popup in den Vordergrund geholt
	//return: true=Popup ist bereits geöffnet, false=Popup existiert nicht
	cancelAll();
	var win=document.getElementById("windowContainer").childNodes;
	for (var t=0;t<win.length;t++) {
		if (win[t].dataset.pagetyp==1 && win[t].dataset.pageid==pageid) {
			focusWindow(win[t].id);
			return true;
		}
	}
	return false;
}

function closePopup(winId) {
	if (winId!="w0") {
		cancelAll();
		//ggf. AutoClose-Timer löschen
		var obj=document.getElementById(winId);
		if (obj) {
			if (obj.dataset.autoclosetimer) {clearTimeout(obj.dataset.autoclosetimer);}
			visu_socket.request_closePopup(obj.dataset.pageid);
		}
		closeWindow(winId);
	}
}

function closeAllPopups() {
	cancelAll();
	var tmp=new Array();
	var win=document.getElementById("windowContainer").childNodes;
	for (var t=0;t<win.length;t++) {
		if (win[t].dataset.pagetyp==1) {tmp.push(win[t].id);}
	}
	
	//alle Popups schliessen (nur DOM)
	for (var t=0;t<tmp.length;t++) {
		if (tmp[t]!="w0") {
			//ggf. AutoClose-Timer löschen
			var obj=document.getElementById(tmp[t]);
			if (obj && obj.dataset.autoclosetimer) {clearTimeout(obj.dataset.autoclosetimer);}
			closeWindow(tmp[t]);
		}
	}
}


/*
============================================================================
Bildschirmschoner
============================================================================
*/

function startScreensaver() {
	if (visu_sstimeout>0 && visu_sspageid>0) {
		screensaverActive=true;
		openPage(visu_sspageid,true);
	}
}

function initScreensaverTimer() {
	//wird bei jeder User-Aktion (Click) von checkClick() aufgerufen, und bei jedem Seitenwechsel (normal und Popup, damit ein Visualarm den Bildschirmschoner deaktiviert)
	screensaverActive=false;
	clearTimeout(screensaverTimer);
	if (visu_sstimeout>0 && visu_sspageid>0) {
		screensaverTimer=window.setTimeout("startScreensaver();",visu_sstimeout*60*1000);
	}
}


/*
============================================================================
Warnungen & Fehlerbehandlung
============================================================================
*/

function showWarnings(arr) {
	var n="";
	if (visu_soundEnabled==1 || visu_textToSpeechEnabled==1) {n+="<div onClick='ackWarningSound(this);' class='visuWarning' style='color:#000000; background:#80e000; pointer-events:auto;'>&#9651;<br>Tonausgabe aktivieren</div>";}	//Achtung: onClick ist wichtig für iOS (nicht downEvent)!
	if (visu_showWarningErr>=0 && parseInt(arr[3])>visu_showWarningErr) {n+="<div onClick='ackWarningErr(this,"+parseInt(arr[3])+");' class='visuWarning' style='color:#ffffff; background:#ff4000; pointer-events:auto;'>&#9651;<br>"+arr[3]+" FEHLER</div>";}	//dto.
	if (arr[0]==1) {n+="<div class='visuWarning'>&#9651;<br>CPU</div>";}
	if (arr[1]==1) {n+="<div class='visuWarning'>&#9651;<br>RAM</div>";}
	if (arr[2]==1) {n+="<div class='visuWarning'>&#9651;<br>HDD</div>";}
	if (n!="") {
		document.getElementById("warn").innerHTML=n;
		document.getElementById("warn").style.display="inline";
	} else {
		document.getElementById("warn").style.display="none";
	}	
}

function ackWarningErr(obj,n) {
	obj.style.display="none";
	visu_showWarningErr=n;
}

function ackWarningSound(obj) {
	visuSoundInit(obj);
	visuTextToSpeechInit(obj);
}

function cancelAll(elementId) {
	//Long-Click
	clearTimeout(clickTimer);
	var obj=document.getElementById("longclick");
	if (obj) {obj.style.display="none";}
	
	//Drag-Operationen ggf. absenden
	if (elementDragId>0) {
		control11_Send();
		control12_Send();
		control13_Send();
		control15_Send();
		elementDragId=0;
	}
	
	//Input-Operationen ggf. abbrechen (bzw. Tastatur ausblenden)
	document.activeElement.blur();
}

function gotoStart() {
	//alles auf Neustart
	cancelAll();
	visu_socket=null;

	//hier alle(!!!) Timer löschen
	//### auch Kameras/Diagramme usw. ###
	clearTimeout(screensaverTimer);
	clearTimeout(errorTimer);

	//alle Fenster entfernen
	clearObject("windowContainer",0);

	//alle Warnungen ausblenden
	document.getElementById("warn").style.display="none";

	//Sounds stoppen
	visuSoundStop();
	visuTextToSpeechStop();

	//wait-Animation einblenden
	document.getElementById("wait").style.display="block";
}

function serverNotReady(visuid) {
	//Server ist nicht bereit (db.sysInfo)
	//d.h. php-Serverscripte starten gerade neu (z.B. beim Compilieren...)
	//=> Visu immer wieder starten, bis es klappt

	gotoStart();

	//kompletter Neustart nach 3 Sekunden
	errorTimer=window.setTimeout("initVisu("+visuid+");",3*1000);
}


/*
============================================================================
Fenster-Verwaltung
============================================================================
*/

function openWindow(xsize,ysize,xpos,ypos) {
	var winID=newWindow();

	if (winID=="w0") { //w0 (Visu-Seite)
		var n="<div id='"+winID+"-page' class='appWindowNormal' style='width:"+xsize+"px; height:"+ysize+"px; box-shadow:none; -webkit-animation:none;'></div>";
		document.getElementById(winID).style.background="none";
	} else { //w1..x (Popups)
		if (xpos!==undefined && ypos!==undefined) {
			n+="<div id='"+winID+"-page' class='appWindowNormal' style='-webkit-animation:none; position:absolute; left:"+xpos+"px; top:"+ypos+"px; width:"+xsize+"px; height:"+ysize+"px; border-radius:3px;'></div>";
		} else {
			var n="<table width='100%' height='100%' cellpadding='0' cellspacing='0' border='0'><tr valign='middle'><td>";
			n+="<div id='"+winID+"-page' class='appWindowNormal' style='width:"+xsize+"px; height:"+ysize+"px; border-radius:3px;'></div>";
			n+="</td></tr></table>";
		}
	}
	document.getElementById(winID).innerHTML=n;
	return winID;
}

function focusWindow(winID) {
	//Popup on top bringen
	if (winID!=null) {
		var win=document.getElementById(winID);

		//neuen winID-zindex = ehemals größer zindex
		var maxZ=0;
		var wins=document.getElementById("windowContainer").childNodes;
		for (var t=0;t<wins.length;t++) {
			if (wins[t].style.zIndex>maxZ) {maxZ=parseInt(wins[t].style.zIndex);}
		}

		//alle zindex, die größer sind als win-zindex: dec()
		for (var t=0;t<wins.length;t++) {
			if (wins[t].style.zIndex>win.style.zIndex) {wins[t].style.zIndex=parseInt(wins[t].style.zIndex)-1;}
		}

		win.style.zIndex=maxZ;
	}
}

function closeWindow(winID) {
	if (winID!=null) {

		var win=document.getElementById(winID);

		//alle zindex, die größer sind als win-zindex: dec()
		var wins=document.getElementById("windowContainer").childNodes;
		for (var t=0;t<wins.length;t++) {
			if (wins[t].style.zIndex>win.style.zIndex) {wins[t].style.zIndex=parseInt(wins[t].style.zIndex)-1;}
		}
		windowZindex--;

		if (win) {win.parentNode.removeChild(win);}
		if (document.getElementById("shortclick")) {document.getElementById("shortclick").style.display="none";}
	}
}

function newWindow() {
	if (document.getElementById("windowContainer").lastChild) {
		var newID=parseInt(document.getElementById("windowContainer").lastChild.id.replace("w",""))+1;
	} else {
		var newID=0;
	}
	var winID="w"+newID;
	var div=document.createElement("div");
	document.getElementById("windowContainer").appendChild(div);
	div.className="appWindow";
	div.id=winID;
	windowZindex++;
	div.style.zIndex=windowZindex;
	return winID;
}


/*
============================================================================
Hilfsfunktionen
============================================================================
*/

function checkClick() {
	//prüft, ob (z.B. auf iOS-Geräten) mehrere Finger ein touch-Event auslösen
	//(wird bei jeder User-Aktion (Click) aufgerufen und setzt daher auch gleich den Bildschirmschoner-Timer zurück)
	//return: true=Singletouch, false=Multitouch
	initScreensaverTimer();
	if (!event) {var event=window.event;}
	if (event && event.touches) {
		if (event.touches.length>1) {
			return false;
		}
	}
	return true;
}

function getPageX(e) {
	//absolute Mouseposition
	if (e.touches) {
		return e.touches[0].pageX;	//e.changedTouches[0].pageX;
	} else {
		return e.pageX;
	}
}

function getPageY(e) {
	//absolute Mouseposition
	if (e.touches) {
		return e.touches[0].pageY;	//e.changedTouches[0].pageY;
	} else {
		return e.pageY;
	}
}

function getClientX(e) {
	//relative Mouseposition (Viewport)
	if (e.touches) {
		return e.touches[0].clientX;	//e.changedTouches[0].clientX;
	} else {
		return e.clientX;
	}
}

function getClientY(e) {
	//relative Mouseposition (Viewport)
	if (e.touches) {
		return e.touches[0].clientY;	//e.changedTouches[0].clientY;
	} else {
		return e.clientY;
	}
}


function disableScrolling() {
	if (!event) var event=window.event;
	event.preventDefault(); //verhindert das Scrollen auf iOS
}

function getDivPosition(obj) {
	//Liefert die absolute Positions eines Elements zurück (klettert quasi den DOM-Baum nach oben)
	//return: Objekt (z.B. Position.x und Position.y)
	var xPos=0;
	var yPos=0;
	while(obj) {
		xPos+=obj.offsetLeft;
		yPos+=obj.offsetTop;
		obj=obj.offsetParent;
	}
	return [xPos,yPos]; //{x:xPos,y:yPos};
}

function createNewDiv(parentId,id) {
	//erzeugt neues Div und hängt es an parent
	//parentId: Parent-Objekt-ID
	//id: gewünschte ID des neuen DIVs
	var div=document.createElement('div');
	document.getElementById(parentId).appendChild(div);
	div.id=id;
	return div;
}

function clickCancel() {
	//verhindert bei verschachtelten DIVs, dass das Click-Event vom Parent bearbeitet wird, wenn auf das Child geklickt wird
	if (!event) var event=window.event;
	if (event) {
		if (event.cancelBubble) {event.cancelBubble=true;}
		if (event.stopPropagation) {event.stopPropagation();}
	}
}

function clearObject(objId,mode) {
	//Leert ein Objekt (entspricht quasi obj.innerHTML="")
	//mode: 1=setzt obj.style.display auf "none"
	var obj=document.getElementById(objId);
	if (mode==1) {obj.style.display="none";}
	while (obj.firstChild) {obj.removeChild(obj.firstChild);}
}

function shakeObj(objId) {
	var obj=document.getElementById(objId);
	if (obj) {
		obj.style.webkitAnimation="none";
		obj.style.webkitAnimationPlayState="paused";
		setTimeout(function(){document.getElementById(objId).style.webkitAnimation="animShake 0.3s linear"; document.getElementById(objId).style.webkitAnimationPlayState="running";},100);
	}
}


/*
============================================================================
Ajax & Websocket
============================================================================
*/

function ajax(cmd,appID,visuID,pageID,data,phpdata) {
	var req=new XMLHttpRequest();
	req.onreadystatechange=function() {
		if (this.readyState==4 && this.status==200) {
			try {
				eval(this.responseText);
			} catch(e) {
				console.log("Ajax-Response fehlerhaft! Error-Msg: "+e.message+" / URL: "+decodeURIComponent(url)+" / Response: "+this.responseText);
//				if (appID==0) { //nur Neustart, wenn grundsätzliche Funktionalität gefährdet ist (also NICHT bei Kamera, Diagramm, etc.)
//					serverNotReady(visuID);
//				}
			}
		}
		if (this.readyState==4 && this.status!=200) {
			console.log("Ajax-Abruf gescheitert! http-status: "+this.status+" / URL: "+decodeURIComponent(url));
			if (appID==0) { //nur Neustart, wenn grundsätzliche Funktionalität gefährdet ist (also NICHT bei Kamera, Diagramm, etc.)
				serverNotReady(visuID);
			}
		}
	}
	//POST
//console.log(data+" /// "+phpdata);
	var url="apps/app"+appID+".php?cmd="+encodeURIComponent(cmd)+"&appid="+encodeURIComponent(appID)+"&visuid="+encodeURIComponent(visuID)+"&pageid="+encodeURIComponent(pageID)+"&sid="+encodeURIComponent(sid)+"&vid="+encodeURIComponent(edomiVersion);
	req.open("POST",url,true);
	req.timeout=30000; //nach 30s: Timeout
	req.setRequestHeader("Content-type","application/x-www-form-urlencoded");
	req.send("data="+encodeURIComponent(data)+"&phpdata="+encodeURIComponent(phpdata));
}

function class_websocket() {
	var that=this;
	
	this.open=function(serverIp,serverPort) {
		try {
			socket=new WebSocket('ws://'+serverIp+':'+serverPort);

			socket.addEventListener('open',function (event) {
				//Browser-Exit => Client offline setzen
				window.onbeforeunload=function (event) {
					if (socket.readyState==1) {
						socket.close();
					}
				}
			});
			
			socket.addEventListener('message',function (event) {
				var tmp=JSON.parse(event.data);
				if (tmp[0]=='LOGIN') 		{that.request_login();}
				else if (tmp[0]=='START') 	{prepareVisu();}
				else if (tmp[0]=='CMD') 	{eval(tmp[1]);}
				else if (tmp[0]=='PING') 	{}
			});
			
			socket.addEventListener('close',function (event) {
//				console.log('Closed '+event.code+'/'+event.reason);
				serverNotReady(visu_visuid);
			});
			
			socket.addEventListener('error',function (event) {
//				console.log('Error ',event.data);
				serverNotReady(visu_visuid);
			});
		
		} catch (exception) {
//			console.log('Exeption '.exception);
			serverNotReady(visu_visuid);
		}
	}

	this.socketSend=function(data) {
		if (socket.readyState==1) {	
			var tmp=JSON.stringify(data);
			socket.send(tmp);
//			console.log("TX: "+tmp);
		}
	}


	this.request_login=function() {
		//Login per SID und visuId
		that.socketSend(['LOGIN',sid,visu_visuid]);
	}

	this.request_initPage=function(pageid) {
		//Seite/Popup aufrufen (KOs updaten)
		that.socketSend(['INITPAGE',sid,pageid]);
	}

	this.request_closePopup=function(pageid) {
		//Popup schließen (KOs updaten)
		that.socketSend(['CLOSEPOPUP',sid,pageid]);
	}

	this.request_execCmdList=function(elementId) {
		//CMD-Liste ausführen
		that.socketSend(['EXECCMDLIST',sid,elementId]);
	}

	this.request_setKoValue=function(koId,value) {
		//KO-Value setzen
		that.socketSend(['SETKOVALUE',sid,koId,value]);
	}


	this.response_initPage=function(page,currentPageIDs,background,item) {
		var renderPage=false;

		if (page.pagetyp==1) {	//Popup
			if (!focusPopup(page.id)) {
				cancelAll();
				clearTimeout(errorTimer);
				renderPage=true;
				if (page.xpos!="" && page.ypos!="") {
					var parentWin=openWindow(page.xsize,page.ysize,page.xpos,page.ypos);
				} else {
					var parentWin=openWindow(page.xsize,page.ysize);
				}
				if (page.autoclose!=0) {
					document.getElementById(parentWin).dataset.autoclosetimer=window.setTimeout(function(){closePopup(parentWin);},page.autoclose*1000);
				}
			}
		} else {	//normale Seite/globale Inkludeseite
			cancelAll();
			clearTimeout(errorTimer);
			renderPage=true;

			closeAllPopups();
			clearObject("w0-page",0);
			var parentWin="w0";
			if (document.getElementById("shortclick")) {document.getElementById("shortclick").style.display="none";}
		}

		if (renderPage) {
			//Page-Daten
			var obj=document.getElementById(parentWin);
			obj.dataset.pageid=page.id;
			obj.dataset.pagetyp=page.pagetyp;

			//Container der Elemente (w#-page): winId (w#) merken (wird für PopupClose gebraucht...)
			document.getElementById(parentWin+"-page").dataset.winid=parentWin;

			//Offset der Seite (Popup) bezogen auf linke/obere Ecke des Browsers
			var pageXpos=document.getElementById(parentWin+"-page").offsetLeft;
			var pageYpos=document.getElementById(parentWin+"-page").offsetTop;
	
			//Seitenhintergrund
			if (background!="") {
				document.getElementById(parentWin+"-page").style.background=background;
				document.getElementById(parentWin+"-page").style.backgroundSize="100% 100%";
				document.getElementById(parentWin+"-page").style.backgroundRepeat="no-repeat";
			} else {
				document.getElementById(parentWin+"-page").style.background="";
			}

			//Visuelemente aufbauen
			for (var t=0;t<item.length;t++) {	
				//DIV erzeugen und allgemeine Eigenschaften
				var obj=createNewDiv(parentWin+"-page","e-"+item[t].id);
				obj.dataset.id=item[t].id;
				obj.dataset.controltyp=item[t].controltyp;

				if (item[t].controltyp==24) {
					//Codeschloss: Länge des Codes
					obj.dataset.var1=item[t].var1.length;
				} else {
					obj.dataset.var1=item[t].var1;
				}
				obj.dataset.var2=item[t].var2;
				obj.dataset.var3=item[t].var3;
				obj.dataset.var4=item[t].var4;
				obj.dataset.var5=item[t].var5;
				obj.dataset.var6=item[t].var6;
				obj.dataset.var7=item[t].var7;
				obj.dataset.var8=item[t].var8;
				obj.dataset.var9=item[t].var9;
				obj.dataset.var10=item[t].var10;
				obj.dataset.gotopageid=item[t].gotopageid;
				obj.dataset.closepopup=item[t].closepopup;
				obj.dataset.hascmd=item[t].hascmd;
				obj.dataset.pagex=pageXpos;
				obj.dataset.pagey=pageYpos;

				//Anklickbar?
				if (item[t].gotopageid>0 || item[t].closepopup>0 || item[t].hascmd>0) {
					obj.dataset.var0="1";
				} else {
					obj.dataset.var0="0";
				}
				
				//GAs
				if (item[t].gaid>0) {
					obj.dataset.gaid=item[t].gaid;
					obj.dataset.value=item[t].kovalue;
				}
				if (item[t].gaid2>0) {
					obj.dataset.gaid2=item[t].gaid2;
					obj.dataset.value2=item[t].kovalue2;
				}

				controlConstruct(obj,item[t]);
				that.refreshItem(item[t].design,true);
			}

		}
	}

	this.response_refreshItems=function(designArr) {
		for (var t=0;t<designArr.length;t++) {
			that.refreshItem(designArr[t],false);
		}
	}
	
	this.refreshItem=function(design,firstInit) {
		var obj=document.getElementById("e-"+design.elementid);
		if (obj) {
			obj.dataset.value=design.kovalue;

			if (design.rcss==1) {
				//Animation mit diesem Trick neustarten (falls es sich um die gleiche Animation wie zuvor handelt, wird die Animation ohne diesen Trick nicht neugestartet!)
				if (design.ranim==1) {
					obj.style.cssText="-webkit-animation:none;";
					obj.offsetWidth;
				}

				//CSS ggf. aus dem Puffer holen bzw. merken (Server überträgt css nur bei Änderung)
				if (design.css!="") {
					obj.style.cssText=pS(design.css,design.kovalue);
					obj.dataset.css=design.css;
				} else {
					obj.style.cssText=pS(obj.dataset.css,design.kovalue);
				}
				
				//Klicks auf unsichtbare Elemente ignorieren
				if (parseFloat(obj.style.opacity)==0) {
					obj.style.pointerEvents="none";
				}
			}

			//Caption ermitteln
			if (design.controltyp==10 || design.controltyp==11 || design.controltyp==13) {
				var tmp=design.text.match(/\{(.*?)\}/);
				if (tmp) {
					design.caption=tmp[0];
				} else {
					design.caption=design.text;
				}
			}
			
			controlRefresh(obj,design,firstInit);
		}
	}

}





/*
============================================================================
Visuelemente & Formulare
============================================================================
*/

function execCmdList(elementId) {
	//Befehlsliste ausführen
	visu_socket.request_execCmdList(elementId);
}

function setKoValue(koId,value) {
	//setzt ein KO auf den Wert value
	visu_socket.request_setKoValue(koId,value);
}

function controlGetFormData(formId) {
	//bildet aus allen Controls in dem "Formular" formId einen String ("1;2;3;...;")
	//die Controls müssen alle eine ID in der Art "x-y-fd#" haben, wobei # eine laufende Nummer ist (fd1,fd2,...)
	//die Reihenfolge ist egal, aus fd1=1,fd3=3,fd2=2 wird "1;2;3;" (NICHT "1;3;2;")
	//fehlende Nummern werden im String aufgefüllt: aus fd1=1,fd2=2,fd4=4 wird "1;2;;4;"
	var n=new Array();
	var tmp=document.getElementById(formId).querySelectorAll("[data-type]");
	for (var t=0; t<tmp.length; t++) {
		var type=tmp[t].dataset.type;
		if (tmp[t].id) {var id=tmp[t].id.split("-")[2].replace("fd","");}
		if (type==1)  {n[id]=stringCleanup(tmp[t].value);}				//Input (Text/Hidden)
		if (type==5)  {n[id]=stringCleanup(tmp[t].dataset.value);}		//Checkbox
		if (type==6)  {n[id]=stringCleanup(tmp[t].dataset.value);}		//Selectbox zum Durchklicken (für TSU)
	}
	var r="";
	for (var t=0; t<n.length; t++) {
		if (n[t]==undefined) {n[t]="";}
		r+=n[t]+AJAX_SEPARATOR1;
	}
	return r;
}

function checkboxInit(obj) {
	if (obj.dataset.value!=0) {
		obj.innerHTML="<button class='elementContent' style='width:100%; height:100%; background:"+visu_indiColor+"; color:"+visu_indiColorText+";'>"+obj.dataset.list.split("|")[1]+"</button>";
	} else {
		obj.innerHTML=obj.dataset.list.split("|")[0];
	}
}

function checkboxToggle(obj) {
	if (obj.dataset.value!=0) {
		obj.dataset.value=0;
	} else {
		obj.dataset.value=1;
	}
	checkboxInit(obj);
}

function selectboxInit(obj) {
	if (obj) {obj.innerHTML=obj.dataset.list.split("|")[obj.dataset.value];}
}

function selectboxToggle(obj) {
	if (obj) {
		var arr=obj.dataset.list.split("|");
		var t=parseInt(obj.dataset.value)+1;
		if (t>=arr.length) {t=0;}
		obj.dataset.value=t;
		selectboxInit(obj);
	}
}

function scrollToTop(objId) {
	var obj=document.getElementById(objId);
	if (obj) {obj.scrollTop=0;}
}

function scrollToEnd(objId) {
	var obj=document.getElementById(objId);
	if (obj) {obj.scrollTop=obj.scrollHeight;}
}

function scrollUp(objId) {
	var obj=document.getElementById(objId);
	if (obj) {obj.scrollTop-=parseInt(obj.clientHeight);}
}

function scrollDown(objId) {
	var obj=document.getElementById(objId);
	if (obj) {obj.scrollTop+=parseInt(obj.clientHeight);}
}

function createWheel(parentObj,wheelObj,wheelScale) {
	//Erzeugt einen "Drehknopf" und passt die Größe und Position an (Verhältnis immer 1:1 und zentriert)
	//parentObj: Container-Objekt (Rad wird entsprechend dimensioniert)
	//wheelObj: das eigentliche Rad-Objekt
	//wheelScale: Anzahl der Pixel, um die das Rad vergrößert/verkleinert werden soll (gegenüber dem parentObj), normalerweise=0

	wheelWidth=3; //Rahmenbreite

	var w=parseFloat(parentObj.style.width.replace("px",""));
	var h=parseFloat(parentObj.style.height.replace("px",""));
	var wheelSize=w;
	if (w>=h) {var wheelSize=h;}
	wheelSize+=wheelScale;

	
	wheelObj.style.border=wheelWidth+"px solid "+visu_indiColor;
	wheelObj.style.borderRadius="100%";

	wheelObj.style.left=(w/2-wheelSize/2)+"px";
	wheelObj.style.top=(h/2-wheelSize/2)+"px";
	wheelObj.style.width=(wheelSize-(2*wheelWidth))+"px"; //-(2*wheelWidth) wegen Border
	wheelObj.style.height=(wheelSize-(2*wheelWidth))+"px";
}

function showShortClickItem(obj) {
	//blendet einen Klick-Indikator für 200 Millisekunden ein
	if (obj) {
		if (!obj.dataset.shortClickTimer) {
			obj.className="shortClick";
		} else {
			clearTimeout(obj.dataset.shortClickTimer);
		}
		obj.dataset.shortClickTimer=window.setTimeout("showShortClickItemReset('"+obj.id+"');",200);
	}
}

function showShortClickItemReset(elementId) {
	var obj=document.getElementById(elementId);
	if (obj) {
		obj.className="";
		obj.dataset.shortClickTimer="";
	}
}

function visuElementClick(elementId,hideClick) {
	var obj=document.getElementById("e-"+elementId);
	if (!hideClick) {showShortClickItem(obj);}

	if (obj.dataset.hascmd>0) { 			//Befehle
		execCmdList(obj.dataset.id);
	}
	if (obj.dataset.gotopageid>0) { 		//Seitenwechsel
		openPage(obj.dataset.gotopageid);
	}
	if (obj.dataset.closepopup>0) { 		//Popup schliessen
		closePopup(obj.parentNode.dataset.winid);
	}

	clickCancel();
}
